from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.tts.tts import MachineBuddy


def str_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    machine_buddy = MachineBuddy()
    print(machine_buddy)
    # machine_buddy.say('hello')

    machine_buddy.change_voice(1)
    machine_buddy.change_rate(250)
    machine_buddy.change_volume(0.5)
    print(machine_buddy)
    # machine_buddy.say('hello')
    return


def instance_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    print('\taudio test')
    machine_buddy = MachineBuddy(
        voice_ind=0,
        rate=150,
        vol=1.0
    )
    machine_buddy.say('instance test')
    return


def static_block_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    print('\taudio test')
    MachineBuddy.speak('static block test', block=True)  # Static use:
    return


def voices_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    machine_buddy = MachineBuddy(rate=150)
    all_voices = machine_buddy.get_all_voices(ack=True)

    print('\taudio test')
    for i, v in enumerate(all_voices):
        machine_buddy.change_voice(new_voice_ind=i)
        machine_buddy.say(text=v.name)
        if 'Hebrew' in str(v.name):
            t = 'שלום, מה קורה חברים?'
            machine_buddy.say(text=t)
    return


def vol_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    machine_buddy = MachineBuddy()
    vol_options = [0.3, 0.6, 0.9]
    print('\taudio test')
    for vol in vol_options:
        machine_buddy.change_volume(new_vol=vol)
        machine_buddy.say(text='volume at {}%'.format(int(vol * 100)))
    return


def rate_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    machine_buddy = MachineBuddy()
    rate_options = [50, 100, 300, 500]
    print('\taudio test')
    for rate in rate_options:
        machine_buddy.change_rate(new_rate=rate)
        machine_buddy.say(text='rate is {}'.format(rate))
    return


def static_without_block_test():
    mt.get_function_name_and_line(ack=True, tabs=0)
    print('\taudio test')
    MachineBuddy.speak('static without block test', block=False)  # Static use:
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name(depth=1)))
    str_test()
    instance_test()
    static_block_test()
    voices_test()
    vol_test()
    rate_test()
    static_without_block_test()
    print('{}'.format('-' * 20))
    return



"""
:requires: cv2
    pip install opencv-python
:sub_package: wizzi_utils.pyplot
"""
try:
    from wizzi_utils.open_cv.open_cv_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.open_cv import test



from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.socket import socket_tools as st
from wizzi_utils.pyplot import pyplot_tools as pyplt
import numpy as np
import os
# noinspection PyPackageRequirements
import cv2

LOOP_TESTS = 50
BLOCK_MS_NORMAL = 2000  # 0 to block
ITERS_CAM_TEST = 10  # 0 to block


def load_img_from_web(name: str, ack: bool = True) -> np.array:
    f = mtt.IMAGES_INPUTS
    url = mtt.IMAGES_D[name]
    suffix = 'jpg'  # default
    # if '.webm' in url:
    #     suffix = 'webm'
    dst = '{}/{}.{}'.format(f, name, suffix)

    if not os.path.exists(dst):
        if not os.path.exists(f):
            mt.create_dir(f)
        success = st.download_file(url, dst)
        if not success:
            mt.exception_error('download failed - creating random img', real_exception=False)
            img = mt.np_random_integers(size=(240, 320, 3), low=0, high=255)
            img = img.astype('uint8')
            cvt.save_img(dst, img)

    img = cvt.load_img(path=dst, ack=ack)
    return img


def get_vid_from_web(name: str) -> str:
    f = mtt.VIDEOS_INPUTS
    url = mtt.VIDEOS_D[name]
    suffix = 'mp4'  # default
    if '.webm' in url:
        suffix = 'webm'
    dst = '{}/{}.{}'.format(f, name, suffix)

    if not os.path.exists(dst):
        if not os.path.exists(f):
            mt.create_dir(f)
        success = st.download_file(url, dst)
        if not success:
            mt.exception_error('download failed - creating random img', real_exception=False)
            dst = None

    return dst


def get_cv_version_test():
    mt.get_function_name(ack=True, tabs=0)
    cvt.get_cv_version(ack=True, tabs=1)
    return


def imread_imwrite_test():
    mt.get_function_name(ack=True, tabs=0)
    name = mtt.SO_LOGO
    img = load_img_from_web(name)

    f = mtt.IMAGES_INPUTS
    url = mtt.IMAGES_D[name]
    dst_path = '{}/{}'.format(f, os.path.basename(url).replace('.png', '_copy.png'))

    cvt.save_img(dst_path, img, ack=True)
    img_loaded = cvt.load_img(dst_path, ack=True)
    print(mt.to_str(img_loaded, '\timg_copy'))
    mt.delete_file(dst_path, ack=True)
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def list_to_cv_image_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    img_list = img.tolist()
    print(mt.to_str(img_list, '\timg_list'))
    img = cvt.list_to_cv_image(img_list)
    print(mt.to_str(img, '\timg'))
    # mt.delete_file(file=mtt.TEMP_IMAGE_PATH, ack=True)
    return


def display_open_cv_image_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    print('\tVisual test: stack overflow logo')
    loc = (70, 200)  # move to X,Y
    resize = 1.7  # enlarge to 170%
    cvt.display_open_cv_image(
        img=img,
        ms=1,  # not blocking
        title='stack overflow logo moved to {} and re-sized to {}'.format(loc, resize),
        loc=loc,  # start from x =70 y = 0
        resize=resize
    )
    loc = pyplt.Location.TOP_RIGHT.value  # move to top right corner
    resize = 1.7  # enlarge to 170%
    cvt.display_open_cv_image(
        img=img,
        ms=BLOCK_MS_NORMAL,  # blocking
        title='stack overflow logo moved to {} and re-sized to {}'.format(loc, resize),
        loc=loc,  # start from x =70 y = 0
        resize=resize
    )
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def display_open_cv_image_loop_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    loc = (70, 200)  # move to X,Y
    resize = 1.7  # enlarge to 170%
    title = 'stack overflow logo moved to {} and re-sized to {} - {} iterations'.format(loc, resize, LOOP_TESTS)
    print('\tVisual test: {}'.format(title))
    for i in range(LOOP_TESTS):
        cvt.display_open_cv_image(
            img=img,
            ms=1,  # not blocking
            title=title,
            loc=loc,  # start from x =70 y = 0
            resize=resize
        )
        if i == 0:  # move just first iter
            loc = None
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def resize_opencv_image_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    print(mt.to_str(img, '\timg'))
    img = cvt.resize_opencv_image(img, scale_percent=0.6)
    print(mt.to_str(img, '\timg re-sized to 60%'))
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def move_cv_img_x_y_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    options = [(0, 0), (100, 0), (0, 100), (150, 150), (400, 400), (250, 350)]
    print('\tVisual test: move to all options {}'.format(options))
    print('\t\tClick Esc to close all')
    for x_y in options:
        title = 'move to ({})'.format(x_y)
        cv2.imshow(title, img)
        cvt.move_cv_img_x_y(title, x_y)
    cv2.waitKey(BLOCK_MS_NORMAL)
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def move_cv_img_by_str_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    options = pyplt.Location.get_location_list_by_rows()
    print('\tVisual test: move to all options {}'.format(options))
    print('\t\tClick Esc to close all')
    for where_to in options:
        title = 'move to {}'.format(where_to)
        cv2.imshow(title, img)
        cvt.move_cv_img_by_str(img, title, where=where_to)
    cv2.waitKey(BLOCK_MS_NORMAL)
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def unpack_list_imgs_to_big_image_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    gray = cvt.BGR_img_to_gray(img)
    big_img = cvt.unpack_list_imgs_to_big_image(
        imgs=[img, gray, img],
        resize=None,
        grid=(2, 2)
    )
    title = 'stack overflow logo 2x2(1 empty)'
    print('\tVisual test: {}'.format(title))
    cvt.display_open_cv_image(
        img=big_img,
        ms=BLOCK_MS_NORMAL,  # blocking
        title=title,
        loc=(0, 0),
        resize=None
    )
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def display_open_cv_images_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    title = '2x1 grid'
    print('\tVisual test: {}'.format(title))
    loc1 = (0, 0)
    cvt.display_open_cv_images(
        imgs=[img, img],
        ms=1,  # blocking
        title='{} loc={}'.format(title, loc1),
        loc=loc1,
        resize=None,
        grid=(2, 1),
        header='{} loc={}'.format(title, loc1),
    )
    loc2 = pyplt.Location.BOTTOM_CENTER.value
    cvt.display_open_cv_images(
        imgs=[img, img],
        ms=BLOCK_MS_NORMAL,  # blocking
        title='{} loc={}'.format(title, loc2),
        loc=loc2,
        resize=None,
        grid=(2, 1),
        header='{} loc={}'.format(title, loc1),
    )
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def display_open_cv_images_loop_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    loc = (70, 200)  # move to X,Y
    title = 'stack overflow logo moved to {} - {} iterations'.format(loc, LOOP_TESTS)
    print('\tVisual test: {}'.format(title))
    for i in range(LOOP_TESTS):
        cvt.display_open_cv_images(
            imgs=[img, img],
            ms=1,  # blocking
            title=title,
            loc=loc,
            resize=None,
            grid=(2, 1),
            header=None
        )
        if i == 0:  # move just first iter
            loc = None
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def gray_to_BGR_and_back_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.SO_LOGO)
    print(mt.to_str(img, '\timgRGB'))
    gray = cvt.BGR_img_to_gray(img)
    print(mt.to_str(img, '\timg_gray'))
    img = cvt.gray_scale_img_to_BGR_form(gray)
    print(mt.to_str(img, '\timgRGB'))
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def BGR_img_to_RGB_and_back_test():
    mt.get_function_name(ack=True, tabs=0)
    imgBGR1 = load_img_from_web(mtt.SO_LOGO)
    print(mt.to_str(imgBGR1, '\timgBGR'))
    imgRGB = cvt.BGR_img_to_RGB(imgBGR1)
    print(mt.to_str(imgRGB, '\timgRGB'))
    imgBGR2 = cvt.RGB_img_to_BGR(imgRGB)
    print(mt.to_str(imgBGR2, '\timgBGR2'))

    cvt.display_open_cv_images(
        imgs=[imgBGR1, imgRGB, imgBGR2],
        ms=BLOCK_MS_NORMAL,  # blocking
        title='imgBGR1, imgRGB, imgBGR2',
        loc=pyplt.Location.CENTER_CENTER,
        resize=None,
        grid=(3, 1),
        header='compare'
    )
    cv2.destroyAllWindows()
    # mt.delete_file(file=mtt.SO_LOGO_PATH, ack=True)
    return


def CameraWu_test(type_cam: str):
    WITH_SLEEP = False
    ports = [0, 1, 13]
    cams = []
    for port in ports:
        cam = cvt.CameraWu.open_camera(port=port, type_cam=type_cam)
        if cam is not None:
            cams.append(cam)

    for cam in cams:
        title = 'CameraWu_test({}) on port {}'.format(cam.type_cam, cam.port)
        fps = mt.FPS(summary_title=title)
        for i in range(ITERS_CAM_TEST):
            fps.start()
            success, cv_img = cam.read_img()
            if WITH_SLEEP:
                mt.sleep(1)

            if success:
                cvt.display_open_cv_image(
                    img=cv_img,
                    ms=1,
                    title=title,
                    loc=pyplt.Location.CENTER_CENTER,
                    resize=None,
                    header='{}/{}'.format(i + 1, ITERS_CAM_TEST)
                )
            fps.update()
        fps.finalize()
    cv2.destroyAllWindows()
    return


def CameraWu_cv2_test():
    mt.get_function_name(ack=True, tabs=0)
    CameraWu_test(type_cam='cv2')
    return


def CameraWu_acapture_test():
    mt.get_function_name(ack=True, tabs=0)
    CameraWu_test(type_cam='acapture')
    return


def CameraWu_imutils_test():
    mt.get_function_name(ack=True, tabs=0)
    CameraWu_test(type_cam='imutils')
    return


def add_text_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.HORSES)
    cvt.add_text(img, header='test text', pos=(100, 100), text_color='r', with_rect=True, bg_color='y', bg_font_scale=2)
    cvt.add_text(img, header='test text', pos=(100, 200), text_color='black', with_rect=True, bg_color='b',
                 bg_font_scale=1)
    cvt.display_open_cv_image(img, ms=BLOCK_MS_NORMAL, loc=pyplt.Location.CENTER_CENTER.value)
    cv2.destroyAllWindows()
    return


def add_header_test():
    mt.get_function_name(ack=True, tabs=0)
    img = load_img_from_web(mtt.HORSES)

    cvt.add_header(img, header='TOP_LEFT', loc=pyplt.Location.TOP_LEFT.value,
                   text_color='lime', with_rect=True, bg_color='azure', bg_font_scale=1)
    cvt.add_header(img, header='BOTTOM_LEFT', loc=pyplt.Location.BOTTOM_LEFT.value,
                   text_color='fuchsia', with_rect=True, bg_color='black', bg_font_scale=2)
    cvt.display_open_cv_image(img, ms=BLOCK_MS_NORMAL, loc=pyplt.Location.CENTER_CENTER.value)
    cv2.destroyAllWindows()

    img = load_img_from_web(mtt.DOG)
    cvt.display_open_cv_image(
        img,
        ms=BLOCK_MS_NORMAL,
        loc=pyplt.Location.CENTER_CENTER.value,
        header='direct header into display_open_cv_image'
    )
    cv2.destroyAllWindows()
    return


def Mp4_creator_test():
    mt.get_function_name(ack=True, tabs=0)
    # now open video file
    vid_name = mtt.DOG1
    video_path = get_vid_from_web(name=vid_name)

    if not os.path.exists(video_path):
        mt.exception_error(mt.NOT_FOUND.format(video_path), real_exception=False)
        return
    cap = cv2.VideoCapture(video_path)
    if cap.isOpened():
        out_dims = cvt.get_dims_from_cap(cap)
        video_total_frames = cvt.get_frames_from_cap(cap)
        print('\tvid {} has {} frames'.format(vid_name, video_total_frames))
        print('\tvid size is {}'.format(out_dims))
    else:
        mt.exception_error('cap is closed.', real_exception=False)
        return

    out_dir = '{}/create_mp4_test'.format(mtt.VIDEOS_OUTPUTS)
    mt.create_dir(out_dir)
    out_fp = '{}/{}_output.mp4'.format(out_dir, vid_name)

    mp4_creator = cvt.Mp4_creator(
        out_full_path=out_fp,
        out_fps=20.0,
        out_dims=out_dims
    )
    print(mp4_creator)

    for i in range(video_total_frames):
        success, frame = cap.read()
        if i % int(video_total_frames / 10) != 0:  # s
            # do only 10 frames
            continue
        print('\tframe {}/{}:'.format(i + 1, video_total_frames))
        # print('\t\t{}'.format(mt.to_str(frame)))
        if success:
            cvt.add_header(
                frame,
                header='create_mp4_test frame {}/{}'.format(i + 1, video_total_frames),
                loc=pyplt.Location.BOTTOM_LEFT.value,
                text_color=pyplt.get_random_color(),
                bg_color=pyplt.get_random_color(),
            )
            cvt.display_open_cv_image(frame, ms=1, title=vid_name, loc=None,
                                      header='{}/{}'.format(i + 1, video_total_frames))
            mp4_creator.add_frame(frame, ack=True, tabs=2)

    cap.release()
    mp4_creator.finalize()
    cv2.destroyAllWindows()
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    get_cv_version_test()
    imread_imwrite_test()
    list_to_cv_image_test()
    display_open_cv_image_test()
    display_open_cv_image_loop_test()
    resize_opencv_image_test()
    move_cv_img_x_y_test()
    move_cv_img_by_str_test()
    unpack_list_imgs_to_big_image_test()
    display_open_cv_images_test()
    display_open_cv_images_loop_test()
    gray_to_BGR_and_back_test()
    BGR_img_to_RGB_and_back_test()
    add_header_test()
    add_text_test()
    CameraWu_cv2_test()
    CameraWu_acapture_test()
    CameraWu_imutils_test()
    Mp4_creator_test()
    print('{}'.format('-' * 20))
    return



from wizzi_utils.json import json_tools as jt
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt


def json_to_string_test():
    mt.get_function_name(ack=True, tabs=0)
    cfg = {'x': 3, 'a': 'dx', 'b': None}
    one_liner = jt.json_to_string(cfg, indent=-1, sort_keys=False)
    print('\tdata in a one liner: {}'.format(one_liner))
    print('\tdata with 4 indentations:')
    print(jt.json_to_string(cfg, indent=4, sort_keys=False, tabs=1))  # pretty print - 4 indentations
    one_liner_sorted = jt.json_to_string(cfg, indent=-1, sort_keys=True)
    print('\tdata in a one liner with sorted keys: {}'.format(one_liner_sorted))
    return


def string_to_json_test():
    mt.get_function_name(ack=True, tabs=0)
    j_str = '{"x": 3, "a": "dx", "b":null}'
    j = jt.string_to_json(j_str)
    one_liner = jt.json_to_string(j, indent=-1, sort_keys=False)
    print('\t{}'.format(one_liner))
    return


def save_load_json_test():
    mt.get_function_name(ack=True, tabs=0)
    j = {'x': 3, 'a': 'dx', 'b': None}
    j2 = {'x2': 39, 'aa': 'dax', 'b7': None}
    mt.create_dir(mtt.TEMP_FOLDER1)
    j_path = '{}/{}.json'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    j2_path = '{}/{}2.json'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    jt.save_json(j_path, j, ack=True)
    jt.save_json(j2_path, j2, ack=True)
    _ = jt.load_json(j_path, ack=True)  # just 1 json file
    _ = jt.load_jsons([j_path, j2_path], ack=True)  # load multiple files
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    json_to_string_test()
    string_to_json_test()
    save_load_json_test()
    print('{}'.format('-' * 20))
    return



from wizzi_utils import misc_tools as mt  # misc tools
import os
import warnings

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # removes tf informative messages
warnings.simplefilter(action='ignore', category=FutureWarning)  # remove tf future warnings
# noinspection PyPackageRequirements,PyUnresolvedReferences
import tensorflow as tf  # noqa E402


def get_tensorflow_version(ack: bool = False, tabs: int = 1) -> str:
    """
    :param ack:
    :param tabs:
    :return:
    see get_tensorflow_version_test()
    """
    # print(tf.__version__)
    # print(tf.VERSION)
    # print(tf.version.VERSION)
    string = mt.add_color('{}* TensorFlow Version {}'.format(tabs * '\t', tf.__version__), ops=mt.SUCCESS_C)
    string += mt.add_color(' - GPU detected ? ', ops=mt.SUCCESS_C)
    if gpu_detected():
        string += mt.add_color('True', ops=mt.SUCCESS_C2)
    else:
        string += mt.add_color('False', ops=mt.FAIL_C2)
    if ack:
        print(string)
    return string


def gpu_detected() -> bool:
    """
    :return:
    """
    gpu_on = False
    try:  # version >= 2
        # noinspection PyUnresolvedReferences
        if len(tf.config.list_physical_devices('GPU')) >= 1:
            gpu_on = True
    except AttributeError:
        if 'GPU' in tf.test.gpu_device_name():
            gpu_on = True
    return gpu_on



from wizzi_utils.tensorflow import tensorflow_tools as tft
from wizzi_utils.misc import misc_tools as mt


def get_tensorflow_version_test():
    mt.get_function_name(ack=True, tabs=0)
    tft.get_tensorflow_version(ack=True)
    return


def gpu_detected_test():
    mt.get_function_name(ack=True, tabs=0)
    print('\tgpu available ? {}'.format(tft.gpu_detected()))
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    get_tensorflow_version_test()
    gpu_detected_test()
    print('{}'.format('-' * 20))
    return



from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.open_cv.test import test_open_cv_tools as cvtt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
# noinspection PyPackageRequirements
import matplotlib.pyplot as plt

SLEEP = 0.2  # sleep between iterations
PLT_MAX_TIME = 2  # close plt after x seconds - minimum 2


def get_colors_formats_test():
    mt.get_function_name(ack=True, tabs=0)
    color_str = 'orange'
    rgb = pyplt.get_RGB_color(color_str)
    rgba = pyplt.get_RGBA_color(color_str)
    bgr = pyplt.get_BGR_color(color_str)
    print('\tcolor {}: RGB={}, RGBA={}, BGR={}'.format(color_str, rgb, rgba, bgr))
    return


def RGBA_to_RGB_and_BGR_test():
    mt.get_function_name(ack=True, tabs=0)
    color_str = 'orange'
    rgb = pyplt.get_RGB_color(color_str)
    bgr = pyplt.get_BGR_color(color_str)

    rgba = pyplt.get_RGBA_color(color_str)
    rgb2 = pyplt.RGBA_to_RGB(rgba)
    bgr2 = pyplt.RGBA_to_BGR(rgba)

    print('\tRGBA {} - {}'.format(rgba, color_str))
    print('\tRGB {}=={} ? {}'.format(rgb, rgb2, rgb == rgb2))
    print('\tBGR {}=={} ? {}'.format(bgr, bgr2, bgr == bgr2))
    return


def BGR_to_RGB_and_RGBA_test():
    mt.get_function_name(ack=True, tabs=0)
    color_str = 'orange'
    rgb = pyplt.get_RGB_color(color_str)
    rgba = pyplt.get_RGBA_color(color_str)

    bgr = pyplt.get_BGR_color(color_str)
    rgb2 = pyplt.BGR_to_RGB(bgr)
    rgba2 = pyplt.BGR_to_RGBA(bgr)
    print('\tBGR {} - {}'.format(bgr, color_str))
    print('\tRGB {}=={} ? {}'.format(rgb, rgb2, rgb == rgb2))
    print('\tRGBA {}=={} ? {}'.format(rgba, rgba2, rgba == rgba2))
    return


def RGB_to_RGBA_and_BGR_test():
    mt.get_function_name(ack=True, tabs=0)
    color_str = 'orange'
    bgr = pyplt.get_BGR_color(color_str)
    rgba = pyplt.get_RGBA_color(color_str)

    rgb = pyplt.get_RGB_color(color_str)
    bgr2 = pyplt.RGB_to_BGR(rgb)
    rgba2 = pyplt.RGB_to_RGBA(rgb)
    print('\tRGB {} - {}'.format(rgb, color_str))
    print('\tRGBA {}=={} ? {}'.format(rgba, rgba2, rgba == rgba2))
    print('\tBGR {}=={} ? {}'.format(bgr, bgr2, bgr == bgr2))
    return


def colors_plot_test():
    mt.get_function_name(ack=True, tabs=0)
    pyplt.colors_plot(ack=True, tabs=0, max_time=PLT_MAX_TIME)
    return


def get_ticks_list_test():
    mt.get_function_name(ack=True, tabs=0)
    ticks_list = pyplt.get_ticks_list(x_low=10, x_high=30, p=0.1)
    print(mt.to_str(ticks_list, '\tticks_list', chars=0))
    ticks_list = pyplt.get_ticks_list(x_low=0, x_high=500, p=0.2)
    print(mt.to_str(ticks_list, '\tticks_list', chars=0))
    return


def get_random_RGBA_color_map_test():
    mt.get_function_name(ack=True, tabs=0)
    random_RBGA_colors = pyplt.get_random_RBGA_color_map(n=3)
    print(mt.to_str(random_RBGA_colors, '\trandom_color_map'))
    return


def screen_dims_test():
    mt.get_function_name(ack=True, tabs=0)
    sd = pyplt.screen_dims()
    print('\tscreen dims {}'.format(sd))
    return


def move_figure_x_y_test():
    mt.get_function_name(ack=True, tabs=0)
    t = np.arange(0.0, 2.0, 0.01)
    s = 1 + np.sin(2 * np.pi * t)
    options = [(0, 0), (100, 0), (0, 100), (150, 150), (400, 400), (250, 350)]
    print('\tVisual test: move to all options {}'.format(options))
    print('\t\tClick Esc to close all')
    timers = []
    for x_y in options:
        title = 'move to {}'.format(x_y)
        fig, ax = plt.subplots()
        pyplt.set_window_title(fig, title='move_figure_x_y_test()')
        pyplt.set_figure_title(fig, title='move_figure_x_y_test()')
        pyplt.set_axes_title(ax, title)
        fig.canvas.mpl_connect('key_press_event', pyplt.on_click_close_all)
        ax.plot(t, s, c=pyplt.get_random_color('str'))
        pyplt.move_figure_x_y(fig, x_y=x_y)
        timer = pyplt.add_timer(fig, PLT_MAX_TIME)
        timers.append(timer)
    plt.show(block=True)
    for timer in timers:
        timer.stop()
    return


def move_figure_by_str_test():
    mt.get_function_name(ack=True, tabs=0)
    t = np.arange(0.0, 2.0, 0.01)
    s = 1 + np.sin(2 * np.pi * t)
    options = pyplt.Location.get_location_list_by_rows()
    print('\tVisual test: move to all options {}'.format(options))
    print('\t\tClick Esc to close all')
    timers = []
    for where_to in options:
        title = 'move to {}'.format(where_to)
        resize = 0.7
        figsize = (6.4 * resize, 4.8 * resize)
        fig, ax = plt.subplots(figsize=figsize)
        pyplt.set_window_title(fig, title=title)
        pyplt.set_figure_title(fig, title=title)
        pyplt.set_axes_title(ax, title)
        fig.canvas.mpl_connect('key_press_event', pyplt.on_click_close_all)
        ax.plot(t, s, c=pyplt.get_random_color('str'))
        pyplt.move_figure_by_str(fig, where=where_to)
        timer = pyplt.add_timer(fig, PLT_MAX_TIME)
        timers.append(timer)
    plt.show(block=True)
    for timer in timers:
        timer.stop()
    return


def add_update_2d_scatter_test():
    mt.get_function_name(ack=True, tabs=0)
    fig, axes = plt.subplots()
    pyplt.set_window_title(fig, 'add_update_2d_scatter_test()')
    # create the fig world on 0 to 4 square
    data = [[0, 0], [4, 4]]  # nx2 base data. if you want a dashboard: data = [0, 0] - just set fig lims
    # c = 'g'  # equivalent: ['g'], pyplt.get_RGBA_color('g'), [pyplt.get_RGBA_color('g')]
    c = ['g', 'r']  # equivalent: [pyplt.get_RGBA_color('g'), 'r']
    sc = pyplt.add_2d_scatter(ax=axes, data=data, c=c, marker='o', marker_s=100, label='sc')
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    new_data = np.array([[2, 2], [3, 3]])  # notice data has to be in the fig world
    # colors = [pyplt.get_RGBA_color('y')]  # list of size 1 or n of RGBA colors
    # colors = pyplt.get_random_RBGA_color_map(n=2)
    colors = [pyplt.get_random_color('RGBA'), pyplt.get_RGBA_color('b')]  # list of size 1 or n of RGBA colors
    pyplt.update_2d_scatter(sc=sc, sc_dict={'data': new_data, 'c': colors})
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    new_data = np.array([[1, 2.33], [3.4, 4.4]])  # notice data has to be in the fig world
    pyplt.update_2d_scatter(sc=sc, sc_dict={'data': new_data})  # no colors
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    pyplt.update_2d_scatter(sc=sc, sc_dict={'data': []})  # no colors

    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    return


def plot_2d_one_figure_iterative_test():
    mt.get_function_name(ack=True, tabs=0)
    iters = 2
    print('\tVisual test: {} iters of plot_2d_iterative_plot_test'.format(iters))

    data_plot_0 = {
        'title': 'main_plot',
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [0, 500, 0, 500],
        'xticks': pyplt.get_ticks_list(x_low=0, x_high=500, p=0.5),
        'datum': [
            {
                'data': mt.np_random_integers(low=0, high=480, size=(20, 2)),
                'c': 'r',
                'label': 'plot0_data0',
                'marker': 'x',
                'marker_s': 1,
            },
            {
                'data': mt.np_random_integers(low=0, high=480, size=(10, 2)),
                'c': pyplt.get_random_color('str'),
                'label': 'plot0_data1',
                'marker': 'o',
                'marker_s': 100,
            }
        ],
    }

    fig, axes, scatters_plot_i = pyplt.plot_2d_one_figure_iterative_init(
        fig_d={
            'title': 'plot_2d_one_figure_iterative_test()',
            'render': True,
            'block': False,
        },
        data_plot_i=data_plot_0,
        win_d={
            'title': 'plot_2d_one_figure_iterative_test()',
            'location': pyplt.Location.CENTER_CENTER.value,
            'resize': 1.2,
            'zoomed': False,
        },
    )

    for i in range(iters):
        mt.sleep(seconds=SLEEP)

        # emulate data of each scatter
        points_per_sc = 20

        new_data_plot_i = []
        for scatter in scatters_plot_i:  # for each scatter in sub plot
            # NOTICE we need same distribution data since the plots are not going to resize the axes
            # noinspection PyProtectedMember
            lows, highs = mt.get_uniform_dist_by_dim(scatter._offsets)
            new_data_i = {
                # 'data': None,  # no refresh
                # 'data': [],  # no refresh
                'data': mt.np_uniform(shape=(points_per_sc, 2), lows=lows, highs=highs),
                # 'c': None,  # color doesn't have to change
                # 'c': [pyplt.get_RGBA_color('g')]  # color could be a list of a color
                'c': pyplt.get_random_RBGA_color_map(points_per_sc)  # color could be a list of n color
            }
            new_data_plot_i.append(new_data_i)

        pyplt.plot_2d_one_figure_iterative_update(
            fig_d={
                'fig': fig,
                'scatters': scatters_plot_i,
                'new_title': 'iter {}'.format(i),
                'render': True,
                'block': False,
            },
            new_data_plot_i=new_data_plot_i,
        )

        # second option if you have more stuff to render
        # pyplt.render_plot(fig, block=block)

    mt.create_dir(mtt.TEMP_FOLDER1)
    save_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    pyplt.save_plot(path=save_path, ack=True, tabs=2)
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def plot_2d_many_figures_iterative_test():
    mt.get_function_name(ack=True, tabs=0)
    iters = 3
    print('\tVisual test: {} iters of plot_2d_iterative_plots'.format(iters))

    data_plot_0 = {
        'title': 'sub plot 0',
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [0, 500, 0, 500],
        'xticks': pyplt.get_ticks_list(x_low=0, x_high=500, p=0.5),
        'yticks': None,
        'datum': [
            {
                'data': mt.np_random_integers(low=0, high=480, size=(20, 2)),
                'c': 'r',
                'label': 'plot0_data0',
                'marker': 'x',
                'marker_s': 100,
            },
            {
                'data': mt.np_random_integers(low=0, high=480, size=(10, 2)),
                'c': pyplt.get_random_color('str'),
                'label': 'plot0_data1',
                'marker': pyplt.get_random_marker(),
                'marker_s': 100,
            }
        ],
    }
    data_plot_1 = {
        'title': 'sub plot 1',
        'legend': {
            'loc': 'lower center',
            'ncol': 1,
            'fancybox': True,
            'framealpha': 0.5,
            'edgecolor': 'black'
        },
        'fig_lims': None,
        'xticks': None,
        'yticks': None,
        'datum': [
            {
                'data': mt.np_uniform(lows=0, highs=0.03, shape=(20, 2)),
                'c': 'magenta',
                'label': 'plot1_data0',
                'marker': '.',
            },
        ],
    }
    data_plot_2 = {  # minimum fields
        'datum': [
            {
                'data': mt.np_normal(mius=0, stds=0.5, shape=(20, 2)),
            },
        ],
    }
    data_plot_3 = {
        'title': 'sub plot 3',
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': None,
        'xticks': None,
        'yticks': None,
        'datum': [
            {
                'data': mt.np_uniform(lows=0, highs=0.03, shape=(20, 2)),
                'c': None,
                'label': 'plot3_data0',
                'marker': None,
            },
        ],
    }

    fig, axes_list, scatters = pyplt.plot_2d_many_figures_iterative_init(
        fig_d={
            'title': 'fig_title',
            'grid_xy': (2, 2),
            'render': True,
            'block': False,
        },
        data_all_plots=[data_plot_0, data_plot_1, data_plot_2, data_plot_3],
        win_d={
            'title': 'plot_2d_many_figures_iterative_test()',
            'location': pyplt.Location.TOP_CENTER.value,
            'resize': 1.5,
            'zoomed': False,
        },
    )

    for i in range(iters):
        mt.sleep(seconds=SLEEP)

        # emulate data of each scatter
        points_per_sc = 20
        new_data_all_plots = []
        for scatters_plot_i in scatters:  # for each sub plot
            new_data_plot_i = []
            for scatter in scatters_plot_i:  # for each scatter in sub plot
                # NOTICE we need same distribution data since the plots are not going to resize the axes
                # noinspection PyProtectedMember
                lows, highs = mt.get_uniform_dist_by_dim(scatter._offsets)
                new_data_i = {
                    # 'data': None,  # no refresh
                    # 'data': [],  # no refresh
                    'data': mt.np_uniform(shape=(points_per_sc, 2), lows=lows, highs=highs),
                    # 'c': None,  # color doesn't have to change
                    # 'c': [pyplt.get_RGBA_color('g')]  # color could be a list of a color
                    'c': pyplt.get_random_RBGA_color_map(points_per_sc)  # color could be a list of n color
                }
                new_data_plot_i.append(new_data_i)
            new_data_all_plots.append(new_data_plot_i)

        pyplt.plot_2d_many_figures_iterative_update(
            fig_d={
                'fig': fig,
                'scatters': scatters,
                'new_title': 'iter {}'.format(i),
                'render': True,
                'block': False,
            },
            new_data_all_plots=new_data_all_plots,
        )

        # second option if you have more stuff to render
        # pyplt.render_plot(fig, block=block)
    mt.create_dir(mtt.TEMP_FOLDER1)
    save_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    pyplt.save_plot(path=save_path, ack=True, tabs=2)
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def plot_2d_one_figure_test():
    mt.get_function_name(ack=True, tabs=0)
    datum = [
        {
            'data': mt.np_random_integers(low=0, high=480, size=(20, 2)),
            'c': pyplt.get_random_color('str'),
            'label': 'plot0_data0',
            'marker': 'x',
        },
        {
            'data': mt.np_random_integers(low=0, high=480, size=(10, 2)),
            'c': 'g',
            'label': 'plot0_data1',
            'marker': 'o',
        }
    ]

    print('\tVisual test: plot_2d_scatter_test')
    pyplt.plot_2d_one_figure(
        datum=datum,
        fig_title='plot_2d_one_figure_test()',
        win_d=pyplt.WINDOW_DEFAULT,
        max_time=PLT_MAX_TIME
    )
    return


def plot_2d_many_figures_test():
    mt.get_function_name(ack=True, tabs=0)
    datum_i = [
        {
            'data': mt.np_random_integers(low=0, high=480, size=(20, 2)),
            'c': 'r',
            'label': 'plot0_data0',
            'marker': 'x',
        },
        {
            'data': mt.np_random_integers(low=0, high=480, size=(10, 2)),
            'c': 'g',
            'label': 'plot0_data1',
            'marker': 'o',
        }
    ]

    datum_list = [datum_i, datum_i]

    print('\tVisual test: plot_2d_scatters_test')
    pyplt.plot_2d_many_figures(
        grid=(1, 2),
        datum_list=datum_list,
        sub_titles=['t1', 't2'],
        fig_title='plot_2d_many_figures_test()',
        win_d=pyplt.WINDOW_DEFAULT,
        max_time=PLT_MAX_TIME
    )
    return


def plot_2d_iterative_dashboard_test():
    mt.get_function_name(ack=True, tabs=0)
    iters = 2
    print('\tVisual test: {} iters of plot_2d_iterative_dashboard_test'.format(iters))

    data_plot_0 = {
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [0, 500, 0, 500],
        'xticks': None,
        'yticks': None,
        'datum': [
            {
                'c': pyplt.get_RGBA_color('r'),
                'label': 'plot0_data0',
                'marker': pyplt.get_random_marker(),
                'marker_s': 125,
            },
            {
                'c': 'g',
                'label': 'plot0_data1',
                'marker': pyplt.get_random_marker(),
            }
        ],
    }

    fig, axes, scatters_plot_i = pyplt.plot_2d_iterative_dashboard_init(
        fig_d={
            'title': 'fig_title',
            'render': True,
            'block': False,
        },
        data_plot_i=data_plot_0,
        win_d={
            'title': 'plot_2d_iterative_dashboard_test()',
            'location': pyplt.Location.BOTTOM_CENTER.value,
            'resize': 1.2,
            'zoomed': False,
        },
        center_d=pyplt.CENTER_DEFAULT,
    )

    for i in range(iters):
        mt.sleep(seconds=SLEEP)

        # emulate data of each scatter
        points_per_sc = 20

        new_data_plot_i = []
        for _ in scatters_plot_i:  # for each scatter in sub plot
            new_data_i = {
                # 'data': None,  # no refresh
                # 'data': [],  # no refresh
                'data': mt.np_random_integers(low=0, high=480, size=(points_per_sc, 2)),
                # 'c': None,  # color doesn't have to change
                # 'c': [pyplt.get_RGBA_color('g')]  # color could be a list of a RGBA color
                # 'c': pyplt.get_random_RBGA_color_map(points_per_sc),  # color could be a list of n RGBA color
            }
            new_data_plot_i.append(new_data_i)

        pyplt.plot_2d_one_figure_iterative_update(
            fig_d={
                'fig': fig,
                'scatters': scatters_plot_i,
                'new_title': 'iter {}'.format(i),
                'render': True,
                'block': False,
            },
            new_data_plot_i=new_data_plot_i,
        )

        # second option if you have more stuff to render
        # pyplt.render_plot(fig, block=block)
    mt.create_dir(mtt.TEMP_FOLDER1)
    save_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    pyplt.save_plot(path=save_path, ack=True, tabs=2)
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def plot_2d_iterative_dashboards_test():
    mt.get_function_name(ack=True, tabs=0)
    iters = 3
    print('\tVisual test: {} iters of plot_2d_iterative_plots'.format(iters))

    data_plot_0 = {
        'title': 'sub plot 0',
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [0, 500, 0, 500],
        'xticks': pyplt.get_ticks_list(x_low=0, x_high=500, p=0.5),
        'datum': [
            {
                'label': 'plot0_data0',
                'marker_s': 120,
            },
            {
                'c': 'g',
                'label': 'plot0_data1',
                'marker': pyplt.get_random_marker(),
            }
        ],
    }
    data_plot_1 = {
        'title': 'sub plot 1',
        'fig_lims': [0, 500, 0, 500],
        'datum': [
            {
                'c': 'magenta',
                'label': 'plot1_data0',
                'marker': pyplt.get_random_marker(),
            },
        ],
    }

    fig, axes_list, scatters = pyplt.plot_2d_iterative_dashboards_init(
        fig_d={
            'title': 'fig_title',
            'grid_xy': (1, 2),
            'render': True,
            'block': False,
        },
        data_all_plots=[data_plot_0, data_plot_1],
        win_d={
            'title': 'plot_2d_iterative_dashboards_test()',
            'location': pyplt.Location.TOP_CENTER.value,
            'resize': 1.5,
            'zoomed': False,
        },
        center_d=pyplt.CENTER_DEFAULT,
    )

    for i in range(iters):
        mt.sleep(seconds=SLEEP)

        # emulate data of each scatter
        points_per_sc = 20
        new_data_all_plots = []
        for scatters_plot_i in scatters:  # for each sub plot
            new_data_plot_i = []
            for _ in scatters_plot_i:  # for each scatter in sub plot
                new_data_i = {
                    # 'data': None,  # no refresh
                    # 'data': [],  # no refresh
                    'data': mt.np_random_integers(low=0, high=480, size=(20, 2)),
                    # 'c': None,  # color doesn't have to change
                    # 'c': [pyplt.get_RGBA_color('g')]  # color could be a list of a color
                    'c': pyplt.get_random_RBGA_color_map(points_per_sc)  # color could be a list of n color
                }
                new_data_plot_i.append(new_data_i)
            new_data_all_plots.append(new_data_plot_i)

        pyplt.plot_2d_many_figures_iterative_update(
            fig_d={
                'fig': fig,
                'scatters': scatters,
                'new_title': 'iter {}'.format(i),
                'render': True,
                'block': False,
            },
            new_data_all_plots=new_data_all_plots,
        )

        # second option if you have more stuff to render
        # pyplt.render_plot(fig, block=block)
    mt.create_dir(mtt.TEMP_FOLDER1)
    save_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    pyplt.save_plot(path=save_path, ack=True, tabs=2)
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def plot_x_y_std_test():
    mt.get_function_name(ack=True, tabs=0)
    data_x = [10, 20, 30]

    C_errors = [5, 7, 1]
    C_errors_stds = [2, 1, 0.5]
    group_c = (C_errors, C_errors_stds, 'g', 'C')

    U_errors = [10, 8, 3]
    U_errors_vars = [4, 3, 1.5]
    group_u = (U_errors, U_errors_vars, pyplt.get_random_color(), 'U')

    print('\tVisual test: errors and stds')
    pyplt.plot_x_y_std(
        data_x,
        groups=[group_c, group_u],
        title='plot_x_y_std_test()',
        legend=pyplt.LEGEND_DEFAULT,
        x_label='X',
        y_label='Y',
        save_path=None,
        show_plot=True,
        with_shift=True,
        max_time=PLT_MAX_TIME
    )
    return


def histogram_test():
    mt.get_function_name(ack=True, tabs=0)
    data = mt.np_uniform(shape=(1000,), lows=0, highs=10000)
    print('\tVisual test: histogram')
    pyplt.histogram(
        values=data,
        title='histogram: 10 bins of 1000 numbers from 0 to 10000',
        save_path=None,
        bins_n=10,
        max_time=PLT_MAX_TIME
    )
    return


def compare_images_sets_test():
    mt.get_function_name(ack=True, tabs=0)
    try:
        print('\tVisual test: 2 compare_images_sets')
        title = 'compare_images_sets_test()'
        imgBGR = cvtt.load_img_from_web(mtt.SO_LOGO)
        imgRGB = cvt.BGR_img_to_RGB(imgBGR)
        img_gray = cvt.BGR_img_to_gray(imgBGR)
        img_gray = cvt.gray_scale_img_to_BGR_form(img_gray)
        # print(mt.to_str(imgBGR))
        # print(mt.to_str(imgRGB))
        # print(mt.to_str(img_gray))
        img_set_a = np.array([imgRGB, imgRGB, imgRGB], dtype=np.uint8)
        img_set_b = np.array([imgRGB, imgBGR, img_gray], dtype=np.uint8)
        # print(mt.to_str(img_set_a))
        # print(mt.to_str(img_set_b))
        pyplt.compare_images_sets(img_set_a, img_set_b, title=title, max_time=PLT_MAX_TIME)
    except ModuleNotFoundError as e:
        mt.exception_error(e, real_exception=True)
    return


def compare_images_multi_sets_squeezed_test():
    mt.get_function_name(ack=True, tabs=0)
    try:
        # # OLD CODE
        # import torch
        # from torchvision import datasets
        # import torchvision.transforms as transforms
        # transform = transforms.Compose([transforms.ToTensor(), ])
        # # choose data set - both work
        # # data_root = path to the data else download
        # print('\tVisual test: 2 compare_images_multi_sets_squeezed')
        # data_root = './Datasets/'
        # mt.create_dir(data_root, ack=True)
        # # dataset = datasets.MNIST(root=data_root, train=False, download=False, transform=transform)
        # dataset = datasets.CIFAR10(root=data_root, train=False, download=False, transform=transform)
        # # noinspection PyUnresolvedReferences
        # data_loader = torch.utils.data.DataLoader(dataset, batch_size=32, shuffle=True, num_workers=2)
        # # noinspection PyUnresolvedReferences
        # images32, labels = iter(data_loader).next()
        #
        # images = images32[:16]  # imagine the first 16 are base images and predicted_images are the model output
        # predicted_images = images32[16:32]
        # print(images.shape)
        # print(predicted_images.shape)
        # d = {'original_data': images, 'predicted_data': predicted_images}
        # pyplt.compare_images_multi_sets_squeezed(
        #     sets_dict=d,
        #     title='comp',
        #     desc=True,
        #     tabs=1
        # )

        print('\tVisual test: 2 compare_images_multi_sets_squeezed')
        title = 'compare_images_multi_sets_squeezed_test()'
        imgBGR = cvtt.load_img_from_web(mtt.SO_LOGO)
        imgRGB = cvt.BGR_img_to_RGB(imgBGR)
        img_gray = cvt.BGR_img_to_gray(imgBGR)
        img_gray = cvt.gray_scale_img_to_BGR_form(img_gray)
        # print(mt.to_str(imgBGR))
        # print(mt.to_str(imgRGB))
        # print(mt.to_str(img_gray))

        # from (height,width,3) -> (3,height,width)
        imgBGR = np.swapaxes(imgBGR, 1, 2)
        imgBGR = np.swapaxes(imgBGR, 0, 1)
        imgRGB = np.swapaxes(imgRGB, 1, 2)
        imgRGB = np.swapaxes(imgRGB, 0, 1)
        img_gray = np.swapaxes(img_gray, 1, 2)
        img_gray = np.swapaxes(img_gray, 0, 1)
        # print(mt.to_str(imgBGR))
        # print(mt.to_str(imgRGB))
        # print(mt.to_str(img_gray))

        # need to normalize by div max color
        img_set_a = np.array([imgRGB, imgRGB, imgRGB], dtype=np.float) / 255
        img_set_b = np.array([imgRGB, imgBGR, img_gray], dtype=np.float) / 255
        # print(mt.to_str(img_set_a))
        # print(mt.to_str(img_set_b))

        d = {
            'original_data': img_set_a,
            'predicted_data': img_set_b,
            'for_fun': img_set_b,
            'for_fun2': img_set_b
        }
        pyplt.compare_images_multi_sets_squeezed(
            sets_dict=d,
            title=title,
            ack=True,
            tabs=1,
            max_time=PLT_MAX_TIME
        )

    except ModuleNotFoundError as e:
        mt.exception_error(e, real_exception=True)
    return


def add_update_3d_scatter_test():
    mt.get_function_name(ack=True, tabs=0)
    fig = plt.figure()
    axes = Axes3D(fig)
    pyplt.set_window_title(fig, 'add_update_2d_scatter_test()')
    # create the fig world on 0 to 5 cube
    # nx3 base data. if you want a dashboard: data = [0, 0,0] - just set fig lims
    data = [[0, 0, 0], [4, 4, 4], [5, 5, 5]]
    # c = 'g'  # equivalent: ['g'], pyplt.get_RGBA_color('g'), [pyplt.get_RGBA_color('g')]
    c = ['g', 'r', 'b']  # equivalent: [pyplt.get_RGBA_color('g'), 'r', 'b]
    sc = pyplt.add_3d_scatter(ax=axes, data=data, c=c, marker='o', marker_s=100, label='sc')
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    new_data = np.array([[2, 2, 2], [3, 3, 3]])  # notice data has to be in the fig world
    # colors = 'y' # equivalent pyplt.get_RGBA_color('y'), [pyplt.get_RGBA_color('y')]
    colors = ['y', pyplt.get_RGBA_color('b')]  # list of size n of RGBA or str colors
    pyplt.update_3d_scatter(sc=sc, sc_dict={'data': new_data, 'c': colors})
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    new_data = np.array([[1, 2.33, 2.1], [3.4, 4.4, 3.1]])  # notice data has to be in the fig world
    pyplt.update_3d_scatter(sc=sc, sc_dict={'data': new_data})  # no colors
    pyplt.render_plot(fig=fig, block=False)

    mt.sleep(seconds=SLEEP)
    pyplt.update_3d_scatter(sc=sc, sc_dict={'data': []})  # no colors

    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    return


def plot_3d_iterative_dashboard_test():
    mt.get_function_name(ack=True, tabs=0)
    iters = 3
    print('\tVisual test: {} iters of 3d iterative plot'.format(iters))

    data_plot_0 = {
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [-20, 20, -20, 20, -20, 20],
        'color_axes': 'b',
        'view': {'azim': -60, 'elev': 30},
        'face_color': 'white',
        'background_color': 'white',
        'xticks': None,
        'yticks': None,
        'zticks': pyplt.get_ticks_list(-20, 20, 0.5),
        'datum': [
            {
                'c': pyplt.get_RGBA_color('r'),
                'label': 'plot0_data0',
                'marker': 'o',
                'marker_s': 150,
            },
            {
                'c': 'g',
                'label': 'plot0_data1',
                'marker': pyplt.get_random_marker(),
                'marker_s': 20,
            }
        ]
    }

    fig, ax, iterative_scatters = pyplt.plot_3d_iterative_dashboard_init(
        fig_d={
            'title': 'iter -1',
            'render': False,
            'block': False,

        },
        data_plot_i=data_plot_0,
        win_d={
            'title': 'plot_3d_iterative_dashboard_test()',
            'location': pyplt.Location.CENTER_CENTER.value,
            'resize': 1.5,
            'zoomed': False,
        },
        center_d=pyplt.CENTER_DEFAULT,
    )

    # # CUSTOM ADD ON 1 - update each round
    center_mass_x_y = {"x1": 0.05, "y1": 0.95, }
    center_mass_label_base = "data0:(xyz)={}"
    center_mass_label = ax.text2D(
        x=center_mass_x_y['x1'],
        y=center_mass_x_y['y1'],
        s=center_mass_label_base.format(np.zeros(3)),
        transform=ax.transAxes,
        color='green'
    )

    # # CUSTOM ADD ON 2 - done once
    pyplt.add_cube3d_around_origin(ax, edge_len=4, add_labels=False)

    # should call legend if you want custom add ons in it
    legend = pyplt.LEGEND_DEFAULT
    plt.legend(loc=legend['loc'], ncol=legend['ncol'], fancybox=legend['fancybox'],
               framealpha=legend['framealpha'], edgecolor=legend['edgecolor'])

    pyplt.render_plot(fig=fig, block=False)  # not necessary - only if you want to see the plot before first data comes
    points_per_sc = 20

    for i in range(iters):
        mt.sleep(seconds=SLEEP)

        # emulate data of each scatter
        new_data_plot_i = []
        for _ in iterative_scatters:  # for each scatter in sub plot
            new_data_i = {
                # 'data': None,  # no refresh
                # 'data': [],  # no refresh
                'data': mt.np_random_integers(low=-15, high=15, size=(points_per_sc, 3)),
                # 'c': None,  # color doesn't have to change
                # 'c': pyplt.get_RGBA_color('r')  # color could be a RGBA color
                # 'c': [pyplt.get_RGBA_color('g')]  # color could be a list of a RGBA color
                'c': pyplt.get_random_RBGA_color_map(points_per_sc),  # color could be a list of n RGBA color
                # 'c': 'r'  # color could be a single str color
                # 'c': 'red'  #  color could be a single str color
                # 'c': ['red'] * points_per_sc  # color could be a list of n string colors
            }
            new_data_plot_i.append(new_data_i)

        pyplt.plot_3d_one_figure_iterative_update(
            fig_d={
                'fig': fig,
                'scatters': iterative_scatters,
                'new_title': 'iter {}'.format(i),
                'render': False,
                'block': False,
            },
            new_data_plot_i=new_data_plot_i,
        )

        # # CUSTOM ADD ON 1 - update each round
        center_mass_label.set_text(
            center_mass_label_base.format(np.round(np.mean(new_data_plot_i[0]['data'], axis=0), 2)))
        pyplt.render_plot(fig, block=False)  # must render here due to custom add on
    mt.create_dir(mtt.TEMP_FOLDER1)
    save_path = '{}/{}'.format(mtt.TEMP_FOLDER1, mtt.JUST_A_NAME)
    pyplt.save_plot(path=save_path, ack=True, tabs=2)
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    mt.delete_dir_with_files(mtt.TEMP_FOLDER1)
    return


def plot_3d_one_dashboard():
    mt.get_function_name(ack=True, tabs=0)
    points_per_sc = 10
    data_plot_0 = {
        'legend': pyplt.LEGEND_DEFAULT,
        'fig_lims': [-20, 20, -20, 20, -20, 20],
        'color_axes': 'b',
        'view': {'azim': -60, 'elev': 30},
        'face_color': 'white',
        'background_color': 'white',
        'xticks': None,
        'yticks': None,
        'zticks': pyplt.get_ticks_list(-20, 20, 0.5),
        'datum': [
            {
                'data': mt.np_random_integers(low=-15, high=15, size=(points_per_sc, 3)),
                'c': 'r',
                'label': 'plot0_data0',
                'marker': 'o',
                'marker_s': 150,
            },
            {
                'data': mt.np_random_integers(low=-15, high=15, size=(points_per_sc, 3)),
                'c': 'g',
                'label': 'plot0_data1',
                'marker': pyplt.get_random_marker(),
                'marker_s': 20,
            }
        ]
    }

    print('\tVisual test: scatter 3d plot')

    pyplt.plot_3d_one_dashboard(
        fig_title='plot_3d_one_dashboard()',
        data_plot_i=data_plot_0,
        win_d={
            'title': 'plot_3d_one_dashboard()',
            'location': pyplt.Location.CENTER_CENTER.value,
            'resize': 1.5,
            'zoomed': False,
        },
        center_d=pyplt.CENTER_DEFAULT,
        max_time=PLT_MAX_TIME
    )

    return


def plot_3d_cube_test():
    mt.get_function_name(ack=True, tabs=0)
    fig = plt.figure()
    pyplt.set_figure_title(fig, title='plot_3d_cube_test()')
    ax = Axes3D(fig)
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    # base on 0,0,0 and get points from left right and top by edge
    edge = 10.0
    point_base = np.array([0., 0., 0.])
    point_edge_left = np.array([0., edge, 0.])
    point_edge_right = np.array([edge, 0., 0.])
    point_edge_top = np.array([0., 0., edge])
    cube_def = [point_base, point_edge_left, point_edge_right, point_edge_top]

    pyplt.plot_3d_cube(
        ax,
        cube_definition=cube_def,
        color='r',
        add_labels=True
    )
    legend = pyplt.LEGEND_DEFAULT
    plt.legend(loc=legend['loc'], ncol=legend['ncol'], fancybox=legend['fancybox'],
               framealpha=legend['framealpha'], edgecolor=legend['edgecolor'])

    print('\tVisual test: plot_3d_cube_test')
    pyplt.move_figure_by_str(plt.gcf())
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    return


def add_cube3d_around_origin_test():
    mt.get_function_name(ack=True, tabs=0)
    fig = plt.figure()
    pyplt.set_figure_title(fig, title='add_cube3d_around_origin_test()')
    ax = Axes3D(fig)
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    pyplt.add_cube3d_around_origin(
        ax,
        edge_len=4,
        color='green',
        add_labels=False
    )

    legend = pyplt.LEGEND_DEFAULT
    plt.legend(loc=legend['loc'], ncol=legend['ncol'], fancybox=legend['fancybox'],
               framealpha=legend['framealpha'], edgecolor=legend['edgecolor'])

    print('\tVisual test: add_cube3d_around_origin')
    pyplt.finalize_plot(fig, PLT_MAX_TIME)
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    get_colors_formats_test()
    RGBA_to_RGB_and_BGR_test()
    BGR_to_RGB_and_RGBA_test()
    RGB_to_RGBA_and_BGR_test()
    get_ticks_list_test()
    colors_plot_test()
    get_random_RGBA_color_map_test()
    screen_dims_test()
    move_figure_x_y_test()
    move_figure_by_str_test()

    # iterative plots - 2d and 3d
    add_update_2d_scatter_test()
    plot_2d_many_figures_iterative_test()
    plot_2d_one_figure_iterative_test()
    plot_2d_iterative_dashboards_test()
    plot_2d_iterative_dashboard_test()
    add_update_3d_scatter_test()
    plot_3d_iterative_dashboard_test()

    # 2d plots
    plot_2d_one_figure_test()
    plot_2d_many_figures_test()
    plot_x_y_std_test()
    histogram_test()
    # 2d comparision
    compare_images_sets_test()
    compare_images_multi_sets_squeezed_test()

    # 3d plots
    plot_3d_one_dashboard()
    plot_3d_cube_test()
    add_cube3d_around_origin_test()

    print('{}'.format('-' * 20))
    return



import numpy as np
import math
import abc
from tflite_runtime.interpreter import Interpreter
# noinspection PyPackageRequirements
import cv2
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.models import BaseModel
from wizzi_utils.models import models_configs as cfg


def get_object_detection_models(ack: bool = False, tabs: int = 1) -> list:
    model_names = []
    count = 0
    for i, (m_name, m_dict) in enumerate(cfg.MODELS_CONFIG.items()):
        if m_dict['job'] == cfg.Jobs.OBJECT_DETECTION.value:
            if ack:
                count += 1
                mt.dict_as_table(table=m_dict, title='{}){}'.format(count, m_name), tabs=tabs)
            model_names.append(m_name)
    return model_names


class OdBaseModel(BaseModel):
    DEFAULT_COLOR_DICT = {
        'bbox': 'r',  # object box
        'label_bbox': 'black',  # label background
        'text': 'white',  # label text
        'sub_image': 'blue',  # label sub image
    }

    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            allowed_class: list = None,
    ):
        super().__init__(save_load_dir=save_load_dir, model_name=model_name)
        if not self.model_name_valid(model_name):
            exit(-1)
        self.model_cfg = cfg.MODELS_CONFIG[model_name]
        self.labels = self.model_cfg['labels_dict']['labels']
        self.allowed_classes = self.labels if allowed_class is None else allowed_class
        return

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def to_string(self, tabs: int = 1) -> str:
        print('abs method - needs implementation')
        exit(-1)
        return ''

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []

    @staticmethod
    def draw_detections(
            detections: list,
            cv_img: np.array,
            colors_d: dict,
            draw_labels: bool,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img() on od models
        :param colors_d: colors in str form:
            bbox color
            label_bbox color
            text color
            if class_id_bbox exist - color the bbox of that class in the given color
            e.g. colors_d={
                    'bbox': 'r',
                    'label_bbox': 'black',
                    'text': 'white',
                    'person_bbox': 'lightgreen', # custom color per class person
                },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_labels: draw label(label_bbox and text on it) on the bbox
        :return:
        """
        for detection in detections:
            label = detection['label']
            score_percentage = detection['score_percentage']
            x0 = detection['bbox']['x0']
            y0 = detection['bbox']['y0']
            x1 = detection['bbox']['x1']
            y1 = detection['bbox']['y1']
            traffic_light_d = detection['traffic_light'] if 'traffic_light' in detection else {}
            bbox_sub_image_d = detection['bbox_sub_image'] if 'bbox_sub_image' in detection else None

            # check custom color for this class
            k = '{}_bbox'.format(label)
            color_bgr = pyplt.get_BGR_color(colors_d[k]) if k in colors_d else pyplt.get_BGR_color(
                colors_d['bbox'])
            cv2.rectangle(img=cv_img, pt1=(x0, y0), pt2=(x1, y1), color=color_bgr, thickness=2)

            for loc, point_and_color in traffic_light_d.items():
                cv2.circle(cv_img, center=tuple(point_and_color['point']), radius=2,
                           color=pyplt.get_BGR_color(point_and_color['color']), thickness=-1)

            if bbox_sub_image_d is not None:
                pt1 = (bbox_sub_image_d['x0'], bbox_sub_image_d['y0'])
                pt2 = (bbox_sub_image_d['x1'], bbox_sub_image_d['y1'])
                cv2.rectangle(img=cv_img, pt1=pt1, pt2=pt2, color=pyplt.get_BGR_color(colors_d['sub_image']),
                              thickness=1)  # sub image

            if draw_labels:
                label_conf = '{}({}%)'.format(label, score_percentage)
                cvt.add_text(cv_img, header=label_conf, pos=(x0, y1), text_color=colors_d['text'], with_rect=True,
                             bg_color=colors_d['label_bbox'], bg_font_scale=2)

        return

    @staticmethod
    def add_traffic_light_to_detections(detections: list, traffic_light_p: dict) -> list:
        """
        :param detections: detections from classify_cv_img()
        :param traffic_light_p: dict of loc to float percentage
            if not none get 3 2d points in a traffic_light form
            e.g. traffic_light={
                    # x of all traffic_light points is frame_width / 2
                    'up': 0.2,  # red light will be take from y = frame_height * 0.2
                    'mid': 0.3,  # yellow light will be take from y = frame_height * 0.3
                    'down': 0.4  # green light will be take from y = frame_height * 0.3
                }
        :return: for each detection in detections: add entry 'traffic_light' with a dict with keys up, mid, down
                    each has location to dict of point and color:
            e.g.
                'traffic_light'= { # x_mid = frame_width / 2
                'up': {'point': [x_mid, y_up], 'color': 'red'},  # y_up = frame_height * traffic_light['up']
                'mid': {'point': [x_mid, y_mid], 'color': 'yellow'},
                'down': {'point': [x_mid, y_down], 'color': 'green'}
                }
        """
        for detection_d in detections:
            # label = detection_d['label']
            # score = detection_d['score_percentage']
            x0 = detection_d['bbox']['x0']
            y0 = detection_d['bbox']['y0']
            x1 = detection_d['bbox']['x1']
            y1 = detection_d['bbox']['y1']
            y_dist = y0 - y1  # image is flipped on y axis
            x_dist = x1 - x0
            # prepare 'traffic_light' points
            # on x middle
            # on y 20% 30% 40% from the top
            x_mid = int(x0 + 0.5 * x_dist)
            y_up = int(y0 - traffic_light_p['up'] * y_dist)
            y_mid = int(y0 - traffic_light_p['mid'] * y_dist)
            y_down = int(y0 - traffic_light_p['down'] * y_dist)
            traffic_light_out = {
                'up': {'point': [x_mid, y_up], 'color': 'red'},
                'mid': {'point': [x_mid, y_mid], 'color': 'yellow'},
                'down': {'point': [x_mid, y_down], 'color': 'green'}
            }
            detection_d['traffic_light'] = traffic_light_out
        return detections

    @staticmethod
    def add_sub_sub_image_to_detection(detections: list, cv_img: np.array, bbox_image_p: dict) -> list:
        """
        :param detections:
        :param cv_img:
        :param bbox_image_p: dict that specify how much from the bbox to save
                bbox_image_p={  # all bbox
                    'x_start': 0,
                    'x_end': 1,
                    'y_start': 0,
                    'y_end': 1,
                },
                bbox_image_p={  # sub bbox
                    'x_start': 0.5,  # start from middle
                    'x_end': 0.9,     # go to the right until 90% of the width
                    'y_start': 0,  # 0 start from the bottom
                    'y_end': 0.4,  # end at 40% from bottom
                },
        :return: for each detection in detections: add entry 'bbox_sub_image' with a dict with keys image,x0,x1,y0,y1
                    if you imshow - you will get the bbox of the detection (according to bbox_image_p)
        """
        for detection_d in detections:
            # label = detection_d['label']
            # confidence = detection_d['score_percentage']
            x0 = detection_d['bbox']['x0']
            y0 = detection_d['bbox']['y0']
            x1 = detection_d['bbox']['x1']
            y1 = detection_d['bbox']['y1']
            y_dist = y0 - y1  # image is flipped on y axis
            x_dist = x1 - x0
            # prepare sub sub image (a part of the bbox)
            # y1 is the bottom
            sub_y0 = int(y1 + bbox_image_p['y_start'] * y_dist)
            sub_y1 = int(y1 + bbox_image_p['y_end'] * y_dist)
            sub_x0 = int(x0 + bbox_image_p['x_start'] * x_dist)
            sub_x1 = int(x0 + bbox_image_p['x_end'] * x_dist)
            detection_d['bbox_sub_image'] = {
                'image': cv_img[sub_y0:sub_y1, sub_x0:sub_x1].tolist(),
                'x0': sub_x0,
                'y0': sub_y0,
                'x1': sub_x1,
                'y1': sub_y1,
            }
        return detections


class TflOdModel(OdBaseModel):
    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            allowed_class: list = None,
            threshold: float = None,
            nms: dict = None,
    ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        :param threshold: only detection above this threshold will be pass first filter
        :param nms: non maximum suppression threshold 2 thresholds
            score_threshold and nms_threshold
            can be None in the cfg - if None pass and None on cfg: no nms
        :param allowed_class: ignore rest of class. list of strings
        example:
        model = od.TflOdModel(
            save_load_dir=m_save_dir,
            model_name=model_name,
        )
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_class=allowed_class)
        if not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.OdTflNormal.value):
            exit(-1)
        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        if threshold is not None:
            self.model_cfg['nms'] = nms

        model_fp = "{}/{}.tflite".format(self.local_path, self.model_name)
        self._download_if_needed(local_path=model_fp, url_dict=self.model_cfg['tflite'])
        self.model_size = mt.file_or_folder_size(model_fp)
        # print('Loading {}(size {}, {} classes)'.format(self.model_name, self.model_size, len(self.labels)))
        self.interpreter = Interpreter(model_path=model_fp, num_threads=4)
        # allocate input output placeholders
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, input_height, input_width, _ = self.interpreter.get_input_details()[0]['shape']
        self.model_cfg['in_dims'] = (input_width, input_height)
        self.model_cfg['input_type'] = self.interpreter.get_input_details()[0]['dtype']
        quantization = self.interpreter.get_input_details()[0]['quantization']
        # need to normalize if model doesn't do quantization (then quantization == (0.0, 0))
        # else e.g. quantization == (0.0078125, 128) - no need to normalize
        self.model_cfg['normalize_RGB'] = True if quantization == (0.0, 0) else False

        # # you can print this to get more details on the model
        # mt.dict_as_table(self.interpreter.get_input_details()[0], title='input')
        # mt.dict_as_table(self.interpreter.get_output_details()[0], title='output')
        # mt.dict_as_table(self.interpreter.get_tensor_details()[0], title='tensor')
        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='TflOdModel:', ops='underlined'))
        string += '\n\t{}name= {} (size {})'.format(tabs_s, self.model_name, self.model_size)
        string += '\n\t{}local_path= {}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_classes, 'allowed_classes'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='cfg', fp=6, ack=False, tabs=tabs + 1))
        return string

    def prepare_input(self, cv_img: np.array) -> np.array:
        """
        :param cv_img:
        resize and change dtype to predefined params
        :return:
        """
        img_RGB = cvt.BGR_img_to_RGB(cv_img)
        if self.model_cfg['normalize_RGB']:
            # normalization is done via the authors of the MobileNet SSD implementation
            center = 127.5
            img_RGB = (img_RGB - center) / center  # normalize image
        img = cv2.resize(img_RGB, self.model_cfg['in_dims'])  # size of this model input
        img_processed = np.expand_dims(img, axis=0).astype(self.model_cfg['input_type'])  # a,b,c -> 1,a,b,c
        return img_processed

    def run_network(self, img_preprocessed: np.array) -> None:
        self.interpreter.set_tensor(self.input_details[0]['index'], img_preprocessed)  # set input tensor
        self.interpreter.invoke()  # run
        return

    def extract_results(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of an object above threshold.
            has items:
                label:str e.g. 'person'
                score_percentage: float e.g. 12.31
                bbox: dict with keys x0,y0,x1,y1
                #  pt1 = (x0, y0)  # obj frame top left corner
                #  pt2 = (x1, y1)  # obj frame bottom right corner
        """
        # get results
        boxes_np = self.interpreter.get_tensor(self.output_details[0]['index'])[0]  # bboxes
        labels_ids_np = self.interpreter.get_tensor(self.output_details[1]['index'])[0]  # labels as list of floats
        scores_np = self.interpreter.get_tensor(self.output_details[2]['index'])[0]  # confidence
        # count_np = interpreter.get_tensor(output_details[3]['index'])[0]  # number of detections

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detections({}) on image {}{}:'.format(self.model_name, len(scores_np), title_suffix,
                                                              cv_img.shape)
            print('{}{}'.format(tabs * '\t', title))
            print('\n{}Meta_data(all detections):'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(labels_ids_np, 'labels_ids_np')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(scores_np, 'scores')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes_np, 'boxes')))

            print('\n{}Detections post first threshold and allowed_class:'.format(tabs * '\t'))
        if len(scores_np) <= 0:
            # no detections made
            if ack:
                print('{}\tNo detections found'.format(tabs * '\t'))
            return []
        boxes_nms_form_list = []  # x,y,w,h in proportion to the image, UNLIKE boxes_np which is x0,y0,x1,y1
        passed_first_filter = []  # threshold <= confidence and label in self.allowed_class
        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        for i, (bbox, label_id, confidence) in enumerate(zip(boxes_np, labels_ids_np, scores_np)):
            if self.model_cfg['threshold'] <= confidence <= 1.0 and label_id is not math.isnan(label_id):
                label = self.labels[int(label_id)]
                if label in self.allowed_classes:
                    # prepare box nms form and save index passed
                    center_x = int(bbox[0] * img_w)
                    center_y = int(bbox[1] * img_h)
                    w = int(bbox[2] * img_w)
                    h = int(bbox[3] * img_h)
                    x = center_x - w / 2
                    y = center_y - h / 2
                    boxes_nms_form_list.append([x, y, w, h])
                    passed_first_filter.append(i)
                    if ack:
                        msg = '{}\t{})Detected class {}: {}({:.2f}%)'
                        print(msg.format(tabs * '\t', i, int(label_id), label, confidence * 100))

        boxes_np = boxes_np[passed_first_filter]
        labels_ids_np = labels_ids_np[passed_first_filter]
        scores_np = scores_np[passed_first_filter]
        if len(scores_np) <= 0:
            if ack:
                print('{}\tNo detections passed first filter'.format(tabs * '\t'))
            return []
        if ack:
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(labels_ids_np, 'labels_ids_np')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(scores_np, 'scores')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes_np, 'boxes')))

        if self.model_cfg['nms'] is not None:  # activate nms
            # https://gist.github.com/leandrobmarinho/26bd5eb9267654dbb9e37f34788486b5
            passed_nms_filter = cv2.dnn.NMSBoxes(
                bboxes=boxes_nms_form_list,
                scores=scores_np.tolist(),
                score_threshold=self.model_cfg['nms']['score_threshold'],
                nms_threshold=self.model_cfg['nms']['nms_threshold']
            )
            if len(passed_nms_filter) > 0:
                passed_nms_filter = passed_nms_filter.flatten()
                boxes_np = boxes_np[passed_nms_filter]
                labels_ids_np = labels_ids_np[passed_nms_filter]
                scores_np = scores_np[passed_nms_filter]
            else:
                scores_np = []

            if ack:
                print('\n{}Meta_data(post-nms):'.format(tabs * '\t'))
                print('{}\t{}'.format(tabs * '\t', mt.to_str(passed_nms_filter, 'pick indices')))
                print('{}\t{}'.format(tabs * '\t', mt.to_str(labels_ids_np, 'labels_ids_np')))
                print('{}\t{}'.format(tabs * '\t', mt.to_str(scores_np, 'scores')))
                print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes_np, 'boxes')))
                print('{}Detections:'.format(tabs * '\t'))
            if len(scores_np) <= 0:
                if ack:
                    print('{}\tNo detections passed nms filter'.format(tabs * '\t'))
                return []

        detections = []
        for i, (bbox, label_id, confidence) in enumerate(zip(boxes_np, labels_ids_np, scores_np)):
            if self.model_cfg['threshold'] <= confidence <= 1.0 and label_id is not math.isnan(label_id):
                label = self.labels[int(label_id)]
                if label in self.allowed_classes:
                    x0 = max(int(bbox[1] * img_w), 0)  # dont exceed 0
                    y0 = max(int(bbox[0] * img_h), 0)  # dont exceed 0
                    x1 = min(int(bbox[3] * img_w), img_w)  # dont exceed frame width
                    y1 = min(int(bbox[2] * img_h), img_h)  # dont exceed frame height
                    score_percentage = round(confidence * 100, fp)

                    detection_d = {
                        'label': label,
                        'score_percentage': score_percentage,
                        'bbox': {
                            #  pt1 = (x0, y0)  # obj frame top left corner
                            #  pt2 = (x1, y1)  # obj frame bottom right corner
                            'x0': x0,
                            'y0': y0,
                            'x1': x1,
                            'y1': y1,
                        },
                    }
                    detections.append(detection_d)
                    if ack:
                        d_msg = '{}\t{})Detected {}({}%) in top left=({}), bottom right=({})'
                        print(d_msg.format(tabs * '\t', i, label, score_percentage, (x0, y0), (x1, y1)))
        return detections

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """

        img_preprocessed = self.prepare_input(cv_img)
        self.run_network(img_preprocessed)
        detections = self.extract_results(
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )
        return detections

    @staticmethod
    def draw_detections(
            detections: list,
            cv_img: np.array,
            colors_d: dict = None,
            draw_labels: bool = True,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img() on od models
        :param colors_d: colors in str form:
            bbox color
            label_bbox color
            text color
            if class_id_bbox exist - color the bbox of that class in the given color
                e.g. colors_d={
                        'bbox': 'r',
                        'label_bbox': 'black',
                        'text': 'white',
                        'person_bbox': 'lightgreen',
                    },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_labels: draw label(label_bbox and text on it) on the bbox
        :return:
        """
        if colors_d is None:
            colors_d = OdBaseModel.DEFAULT_COLOR_DICT
        OdBaseModel.draw_detections(
            detections=detections,
            cv_img=cv_img,
            colors_d=colors_d,
            draw_labels=draw_labels
        )
        return

    @staticmethod
    def add_traffic_light_to_detections(detections: list, traffic_light_p: dict) -> list:
        """
        see add_traffic_light_to_detections
        """
        detections = OdBaseModel.add_traffic_light_to_detections(
            detections=detections,
            traffic_light_p=traffic_light_p
        )
        return detections

    @staticmethod
    def add_sub_sub_image_to_detection(detections: list, cv_img: np.array, bbox_image_p: dict) -> list:
        """
        see add_sub_sub_image_to_detection
        """
        detections = OdBaseModel.add_sub_sub_image_to_detection(
            detections=detections,
            cv_img=cv_img,
            bbox_image_p=bbox_image_p
        )
        return detections


class Cv2OdModel(OdBaseModel):
    def __init__(self,
                 save_load_dir: str,
                 model_name: str,
                 allowed_class: list = None,
                 threshold: float = None,
                 nms_threshold: float = None,
                 in_dims: tuple = None,
                 scalefactor: float = None,
                 mean: tuple = None,
                 swapRB: bool = None,
                 crop: bool = None,
                 ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        :param allowed_class: ignore rest of class. list of strings
            if None - all classes allowed
        for all the following - if is None: take default value from MODELS_DNN_OBJECT_DETECTION_META_DATA['model_name']
        :param threshold: only detection above this threshold will be returned
        :param nms_threshold: non maximum suppression threshold
        :param in_dims:
        :param scalefactor:
        :param mean:
        :param swapRB:
        :param crop:
        example:
        model = od.Cv2OdModel(
            save_load_dir=m_save_dir,
            model_name=model_name,
            allowed_class=['dog', 'cat'],
            threshold=0.1,
            nms_threshold=0.3,
            in_dims=(416, 416),
            scalefactor=1 / 127.5,
            mean=(0, 0, 0),
            swapRB=True,
            crop=False,
        )
        see:
        best_model_images_test()
        best_model_video_test()
        best_model_cam_test()
        models_compare_images_test()
        models_compare_video_test()
        models_compare_cam_test()
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_class=allowed_class)
        if not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.OdCvNormal.value):
            exit(-1)
        net = None
        if self.model_cfg['family'] == cfg.DnnFamily.Caffe.value:
            model_prototxt = "{}/{}.prototxt".format(self.local_path, self.model_name)
            model_caffe = "{}/{}.caffemodel".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_prototxt, url_dict=self.model_cfg['prototxt'])
            self._download_if_needed(local_path=model_caffe, url_dict=self.model_cfg['caffemodel'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_prototxt), mt.file_or_folder_size(model_caffe)]
            net = cv2.dnn.readNetFromCaffe(prototxt=model_prototxt, caffeModel=model_caffe)

        elif self.model_cfg['family'] == cfg.DnnFamily.Darknet.value:
            model_cfg = "{}/{}.cfg".format(self.local_path, self.model_name)
            model_weights = "{}/{}.weights".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_cfg, url_dict=self.model_cfg['cfg'])
            self._download_if_needed(local_path=model_weights, url_dict=self.model_cfg['weights'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_cfg), mt.file_or_folder_size(model_weights)]
            net = cv2.dnn.readNetFromDarknet(cfgFile=model_cfg, darknetModel=model_weights)

        elif self.model_cfg['family'] == cfg.DnnFamily.TF.value:
            model_pbtxt = "{}/{}.pbtxt".format(self.local_path, self.model_name)
            model_pb = "{}/{}.pb".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_pbtxt, url_dict=self.model_cfg['pbtxt'])
            self._download_if_needed(local_path=model_pb, url_dict=self.model_cfg['pb'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_pbtxt), mt.file_or_folder_size(model_pb)]
            net = cv2.dnn.readNetFromTensorflow(model=model_pb, config=model_pbtxt)

        if net is None:
            mt.exception_error('Failed to create network', real_exception=False)
            exit(-1)

        self.model = cv2.dnn_DetectionModel(net)

        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        if nms_threshold is not None:
            self.model_cfg['nms_threshold'] = nms_threshold
        if in_dims is not None:
            self.model_cfg['in_dims'] = in_dims
        if scalefactor is not None:
            self.model_cfg['scalefactor'] = scalefactor
        if mean is not None:
            self.model_cfg['mean'] = mean
        if swapRB is not None:
            self.model_cfg['swapRB'] = swapRB
        if crop is not None:
            self.model_cfg['crop'] = crop

        # self.model.setInputParams(size=(416, 416), scale=1 / 255, swapRB=True)
        self.model.setInputParams(
            scale=self.model_cfg['scalefactor'],
            size=self.model_cfg['in_dims'],
            mean=self.model_cfg['mean'],
            swapRB=self.model_cfg['swapRB'],
            crop=self.model_cfg['crop']
        )

        # TODO test GPU with cuda 11 installed
        # count = cv2.cuda.getCudaEnabledDeviceCount()
        # print(count)
        # self.network.setPreferableBackend(cv2.dnn.DNN_BACKEND_CUDA)
        # self.network.setPreferableTarget(cv2.dnn.DNN_TARGET_CUDA)
        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='Cv2OdModel:', ops='underlined'))
        string += '\n\t{}name= {} (resources size: {})'.format(tabs_s, self.model_name, self.model_resources_sizes)
        string += '\n\t{}local_path={}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_classes, 'allowed_class'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='conf', fp=6, ack=False, tabs=tabs + 1))
        return string

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        classes, scores, boxes = self.model.detect(cv_img, self.model_cfg['threshold'], self.model_cfg['nms_threshold'])
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(classes, 'classes')))
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(scores, 'scores')))
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes, 'boxes')))
        if len(classes) > 0 and len(scores) > 0:
            classes = classes.flatten()
            scores = scores.flatten()

        detections = self.extract_results(
            classes=classes,
            scores=scores,
            boxes=boxes,
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )

        return detections

    def extract_results(
            self,
            classes: np.array,
            scores: np.array,
            boxes: np.array,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param classes: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param scores: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param boxes: result of classes, scores, boxes = self.model.detect(cv_img, self.threshold, self.nms_threshold)
        :param cv_img: cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of an object above threshold.
            has items:
                label:str e.g. 'person'
                score_percentage: float e.g. 12.31
                bbox: dict with keys x0,y0,x1,y1
                #  pt1 = (x0, y0)  # obj frame top left corner
                #  pt2 = (x1, y1)  # obj frame bottom right corner
        """

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detections({}) on image {}{}:'.format(self.model_name, len(classes), title_suffix, cv_img.shape)

            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(classes, 'classes')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(scores, 'scores')))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(boxes, 'boxes')))
            print('{}Detections:'.format(tabs * '\t'))

        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        detections = []
        for i, (class_id, score_frac, bbox) in enumerate(zip(classes, scores, boxes)):
            label = self.labels[class_id]
            if label in self.allowed_classes:
                score_percentage = round(score_frac * 100, fp)

                (x0, y0) = (bbox[0], bbox[1])
                (w, h) = (bbox[2], bbox[3])

                x1 = min(x0 + w, img_w)  # dont exceed frame width
                y1 = min(y0 + h, img_h)  # dont exceed frame height

                if 'need_normalize' in self.model_cfg:
                    in_dims_w, in_dims_h = self.model_cfg['in_dims']
                    x0 = int(x0 * (img_w / in_dims_w))
                    x1 = int(x1 * (img_w / in_dims_w))
                    y0 = int(y0 * (img_h / in_dims_h))
                    y1 = int(y1 * (img_h / in_dims_h))

                detection_d = {
                    'label': label,
                    'score_percentage': score_percentage,
                    'bbox': {
                        #  pt1 = (x0, y0)  # obj frame top left corner
                        #  pt2 = (x1, y1)  # obj frame bottom right corner
                        'x0': int(x0),
                        'y0': int(y0),
                        'x1': int(x1),
                        'y1': int(y1),
                    },
                }

                detections.append(detection_d)
                if ack:
                    d_msg = '{}\t{})Detected {}({}%) in top left=({}), bottom right=({})'
                    print(d_msg.format(tabs * '\t', i, label, score_percentage, (x0, y0), (x1, y1)))

        return detections

    @staticmethod
    def draw_detections(
            detections: list,
            cv_img: np.array,
            colors_d: dict = None,
            draw_labels: bool = True,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img() on od models
        :param colors_d: colors in str form:
            bbox color
            label_bbox color
            text color
            if class_id_bbox exist - color the bbox of that class in the given color
                e.g. colors_d={
                        'bbox': 'r',
                        'label_bbox': 'black',
                        'text': 'white',
                        'person_bbox': 'lightgreen',
                    },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_labels: draw label(label_bbox and text on it) on the bbox
        :return:
        """
        if colors_d is None:
            colors_d = OdBaseModel.DEFAULT_COLOR_DICT
        OdBaseModel.draw_detections(
            detections=detections,
            cv_img=cv_img,
            colors_d=colors_d,
            draw_labels=draw_labels
        )
        return

    @staticmethod
    def add_traffic_light_to_detections(detections: list, traffic_light_p: dict) -> list:
        """
        see add_traffic_light_to_detections
        """
        detections = OdBaseModel.add_traffic_light_to_detections(
            detections=detections,
            traffic_light_p=traffic_light_p
        )
        return detections

    @staticmethod
    def add_sub_sub_image_to_detection(detections: list, cv_img: np.array, bbox_image_p: dict) -> list:
        """
        see add_sub_sub_image_to_detection
        """
        detections = OdBaseModel.add_sub_sub_image_to_detection(
            detections=detections,
            cv_img=cv_img,
            bbox_image_p=bbox_image_p
        )
        return detections



import numpy as np
from wizzi_utils import misc_tools as mt


def find_centers(A: np.array, k: int = 1) -> np.array:
    """
    :requires: sklearn
    :param A: nx(d+1) data array. A=X|y
    :param k: how many centers
    :return: centers kxd
    see find_centers_test()

    if you get No module named 'sklearn.__check_build._check_build'
    pip uninstall scikit-learn
    pip uninstall sklearn
    pip install sklearn
    """
    try:
        # noinspection PyPackageRequirements
        from sklearn.cluster import KMeans
        k_means_obj = KMeans(n_clusters=k)
        k_means_obj.fit(A)
        centers = k_means_obj.cluster_centers_
    except (ModuleNotFoundError, ImportError) as e:
        centers = None
        mt.exception_error(e)
    return centers
